package tcc.haruki.domain.xtr;

import java.math.BigInteger;

import tcc.haruki.domain.XTRElem;
import tcc.haruki.domain.GaloisField;
import tcc.haruki.domain.XTRTrincaElem;
import tcc.haruki.exceptions.XTRInvalidParametersException;
import tcc.haruki.exceptions.XTRInvalidKeyException;
import tcc.haruki.exceptions.XTRNullParametersException;
import tcc.haruki.services.XTROperationService;
import tcc.haruki.services.impl.XTRDefaultOperationService;

public class XTRKey {
	private XTRPrivateKey privateKey;
	private XTRPublicKey publicKey;
	
	private XTRKey (XTRPrivateKey privateKey, XTRPublicKey publicKey) {
		this.privateKey = privateKey;
		this.publicKey = publicKey;
	}

	/**
	 * Gera um Tr(g) pertencente a GF(p<sup>2</sup>) com ordem <code>q</code>
	 * @param gf
	 * @param q
	 * @return Tr(g) de ordem q
	 * @throws XTRNullParametersException 
	 */
	public static XTRElem generateTraceOrderQ(XTRParameters params) throws XTRNullParametersException {
		XTRElem c;
		XTRElem cpM1;
		XTRElem cbla;
		
		GaloisField gf = new GaloisField(params.getP());
		
		XTROperationService service = new XTRDefaultOperationService(params);
		
		do {
			do {
				do {
					c =	new XTRElem(gf, gf.randomElement(), gf.randomElement());
				} while (c.isGFElement());
				
				XTRTrincaElem spM1 = service.Sn(c, gf.getP().add(BigInteger.ONE));
				cpM1 = spM1.get(0);
			} while (cpM1.isGFElement());
			
			XTRTrincaElem sbla = service.Sn(c, gf.getP().multiply(gf.getP()) .subtract(gf.getP()) .add(BigInteger.ONE) .divide(params.getQ()));
			cbla = sbla.get(0);
		} while (cbla.equals(new XTRElem(gf, new BigInteger("-3"), new BigInteger("-3"))));
		
		return cbla;
	}
	
	public static XTRKey generateXTRKey() throws XTRNullParametersException, XTRInvalidParametersException, XTRInvalidKeyException {
		return generateXTRKey(XTRParameters.generateParameters());
	}
	
	public static XTRKey generateXTRKey(XTRParameters params) throws XTRNullParametersException, XTRInvalidParametersException, XTRInvalidKeyException {
		XTRElem trg = generateTraceOrderQ(params);
		
		XTROperationService service = new XTRDefaultOperationService(params);
		BigInteger k = service.random1AtehQ();
		
		XTRPrivateKey privateKey = new XTRPrivateKey(params, trg, k);
		XTRPublicKey publicKey = new XTRPublicKey(params, trg, service.Sn(trg, k));
		
		XTRKey key = new XTRKey(privateKey, publicKey);
		return key;
	}
	
	public static XTRKey initializeKey(XTRPrivateKey privateKey, XTRPublicKey publicKey) throws XTRInvalidKeyException{
		XTRKey key = new XTRKey(privateKey, publicKey);
		if (!key.isValid()) {
			throw new XTRInvalidKeyException();
		}
		return key;
	}


	public XTRPrivateKey getPrivateKey() {
		return privateKey;
	}


	public XTRPublicKey getPublicKey() {
		return publicKey;
	}
	
	public boolean isValid() {
		if (privateKey == null || publicKey == null) {
			return false;
		}
		if (privateKey.getP() != publicKey.getP() || privateKey.getQ() != publicKey.getQ() || 
				privateKey.getTrg() != publicKey.getTrg()) {
			return false;
		}
		try {
			XTROperationService service = new XTRDefaultOperationService(XTRParameters.makeParametersWith(privateKey.getP(), privateKey.getQ()));
			if (! service.Sn(privateKey.getTrg(), privateKey.getK()).equals(publicKey.getH())) {
				return false;
			}
		} catch (XTRNullParametersException e) {
			return false;
		} catch (XTRInvalidParametersException e) {
			return false;
		}
		
		return true;
	}
}
